<?php

namespace Drupal\commerce_ups_fulfillment\Form;

use Drupal\Core\Form\ConfigFormBase;
use Drupal\Core\Form\FormStateInterface;
use Drupal\commerce_shipping\Entity\ShippingMethod;

/**
 * Defines a form to configure UPS Fulfillment settings.
 */
class UPSFulfillmentConfigForm extends ConfigFormBase {

  /**
   * {@inheritdoc}
   */
  public function getFormId() {
    return 'commerce_ups_fulfillment_config_form';
  }

  /**
   * {@inheritdoc}
   */
  protected function getEditableConfigNames() {
    return ['commerce_ups_fulfillment.ups_fulfillment_settings'];
  }

  /**
   * {@inheritdoc}
   */
  public function buildForm(array $form, FormStateInterface $form_state) {
    $config = $this->config('commerce_ups_fulfillment.ups_fulfillment_settings');

    // Load all shipping methods.
    $shipping_methods = ShippingMethod::loadMultiple();
    $ups_plugin_count = 0;
    $ups_in_test_mode = FALSE;
    $missing_client_secret = FALSE;

    foreach ($shipping_methods as $method_id => $method) {

      if (!$method->isEnabled()) {
        continue;
      }

      $plugin = $method->getPlugin();

      // Check if the plugin ID is 'ups'.
      if ($plugin->getPluginId() === 'ups') {
        $ups_plugin_count++;

        // Check if the UPS shipping method is in "test" mode.
        if ($plugin->getConfiguration()['api_information']['mode'] === 'test') {
          $ups_in_test_mode = TRUE;
        }

        // Check if the client secret is missing.
        if (empty($plugin->getConfiguration()['api_information']['secret'])) {
          $missing_client_secret = TRUE;
        }
      }
    }

    // If no enabled plugins are found, search for disabled plugins (GTCC workaround)
    if ($ups_plugin_count === 0) {
      foreach ($shipping_methods as $method_id => $method) {

        $plugin = $method->getPlugin();

        // Check if the plugin ID is 'ups'.
        if ($plugin->getPluginId() === 'ups') {
          $ups_plugin_count++;

          // Check if the UPS shipping method is in "test" mode.
          if ($plugin->getConfiguration()['api_information']['mode'] === 'test') {
            $ups_in_test_mode = TRUE;
          }

          // Check if the client secret is missing.
          if (empty($plugin->getConfiguration()['api_information']['secret'])) {
            $missing_client_secret = TRUE;
          }
        }
      }
    }

    // If no 'ups' plugin or multiple 'ups' plugins are found, show an error.
    if ($ups_plugin_count === 0) {
      $this->messenger()->addError($this->t('No UPS shipping method found.'));
    } elseif ($ups_plugin_count > 1) {
      $this->messenger()->addError($this->t('Multiple UPS shipping methods found. Only one at a time is supported.'));
    }

    // If the UPS method is in "test" mode, display a warning.
    if ($ups_in_test_mode) {
      $this->messenger()->addWarning($this->t('The UPS shipping method is currently in test mode.'));
    }

    // If the client secret is missing, display an error.
    if ($missing_client_secret) {
      $this->messenger()->addError($this->t('The UPS shipping method is missing the client secret.'));
    }

    // If there are no issues, proceed with building the form.
    if ($ups_plugin_count === 1 && !$missing_client_secret) {

      $form['ups_fulfillment_enabled'] = [
        '#type' => 'checkbox',
        '#title' => $this->t('UPS Fulfillment Enabled'),
        '#default_value' => $config->get('ups_fulfillment_enabled'),
        '#ajax' => [
          'callback' => '::toggleShippingMethodCheckboxes',
        ],
      ];

      $form['shipping_methods_details'] = [
        '#type' => 'details',
        '#title' => $this->t('Enable for Shipping Methods'),
        '#open' => TRUE,
        '#states' => [
          'visible' => [
            ':input[name="ups_fulfillment_enabled"]' => ['checked' => TRUE],
          ],
        ],
      ];

      // Container for the shipping methods checkboxes.
      $form['shipping_methods_details']['shipping_methods_container'] = [
        '#type' => 'container',
        '#attributes' => ['id' => 'shipping-methods-wrapper'],
      ];

      $enabled_methods = $config->get('enabled_shipping_methods') ?? [];

      // Add a checkbox for the valid UPS shipping method.
      foreach ($shipping_methods as $method_id => $method) {
        $form['shipping_methods_details']['shipping_methods_container'][$method_id] = [
          '#type' => 'checkbox',
          '#title' => $method->label(),
          '#default_value' => in_array($method_id, $enabled_methods),
        ];
      }
    }

    return parent::buildForm($form, $form_state);
  }


  /**
   * Ajax callback to refresh the shipping methods checkboxes.
   */
  public function toggleShippingMethodCheckboxes(array &$form, FormStateInterface $form_state) {
    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function submitForm(array &$form, FormStateInterface $form_state) {

    $this->config('commerce_ups_fulfillment.ups_fulfillment_settings')
      ->set('ups_fulfillment_enabled', $form_state->getValue('ups_fulfillment_enabled'))
      ->set('enabled_shipping_methods', array_keys(array_filter($form_state->getValues('shipping_methods_container'))))
      ->save();

    parent::submitForm($form, $form_state);
  }
}
