<?php

namespace Drupal\commerce_ups_fulfillment;

use Drupal\commerce_price\Price;
use Drupal\commerce_shipping\Entity\Shipment;
use Drupal\commerce_ups\UPSSdk;
use Drupal\Component\Serialization\Json;
use Drupal\physical\LengthUnit;
use Drupal\physical\WeightUnit;
use Drupal\Core\File\FileSystemInterface;

class UPSFulfillmentSdk extends UPSSdk {

  /**
   * Url for Shipping API request.
   */
  const UPS_API_SHIPPING_URL = '/api/shipments/v1/ship';

  /**
   * Url for Label Recovery API request.
   */
  const UPS_API_LABEL_RECOVERY_URL = '/api/labels/v1/recovery';

  /**
   * Url for Void Shipping API request.
   */
  const UPS_API_VOID_SHIPPING_URL = '/api/shipments/v1//void/cancel/';

  /**
   * Creates a new UPS Shipment
   */
  public function createShipment(): array {
    if (!$this->shipment) {
      throw new \InvalidArgumentException('Shipment not provided.');
    }
    $params = $this->prepareShipmentRequest();

    if (empty($params)) {
      \Drupal::logger('commerce_ups_fulfillment')->error('Failed to prepare request parameters.');
      return [];
    }

    if ($this->logRequest) {
      $this->log('Sending UPS Shipment request with params', $params);
    }

    try{

      $response = $this->client->post(self::UPS_API_SHIPPING_URL, [
        'json' => $params,
        'headers' => ['Content-Type' => 'application/json;charset=UTF-8', 'Accept' => '*/*',],
        'debug' => true ]
      );

      $shipment = Json::decode($response->getBody());
      
      if ($this->logResponse) {
        $this->log("Received UPS Shipment response", $shipment);
      }
      if (!empty($shipment["ShipmentResponse"]["ShipmentResults"])) {

        $totalCharges = $shipment["ShipmentResponse"]["ShipmentResults"]['ShipmentCharges']['TotalCharges'];
        $shipmentAmount = new Price($totalCharges['MonetaryValue'], $totalCharges['CurrencyCode']);

        if ($this->shipment->hasField('field_remote_price')) {
          $this->shipment->set('field_remote_price', $shipmentAmount);
        }
        $this->shipment->setAmount($shipmentAmount);

        $tracking_number = $shipment["ShipmentResponse"]["ShipmentResults"]["PackageResults"]["TrackingNumber"];
        $gif_data = $shipment["ShipmentResponse"]["ShipmentResults"]["PackageResults"]["ShippingLabel"]["GraphicImage"];
        $file_name = $tracking_number . '.gif';
        $this->shipment->setTrackingCode($tracking_number);

        // Shipment is saved in this function
        $this->attach_image_to_shipment($this->shipment, $gif_data, $file_name);

        return $shipment;
      }
    } catch (\GuzzleHttp\Exception\RequestException $e) {
      // An error occurred while sending the shipment to UPS
      // Display this error to the user, it's likely a missing phone number, etc.
      $response = $e->getResponse();
      $message = 'An error occurred while communicating with UPS.';

      if ($response) {
        $response_body = $response->getBody()->getContents();
        $decoded_response = Json::decode($response_body, TRUE);

        if (isset($decoded_response['response']['errors'])) {
          $errors = $decoded_response['response']['errors'];
          if (is_array($errors)) {
            // UPS Errors have a "Code" and "Message"
            $formatted_errors = array_map(function ($error) {
              return "Code: {$error['code']}, Message: {$error['message']}";
            }, $errors);
            $message .= ' - ' . implode('; ', $formatted_errors);
          } else {
            // Response was not an array, display it as a string
            $message .= ' - ' . $errors;
          }
        } else {
          // Error response was not in the expected format, attempt to display it
          $message .= ' - Response - ' . $response_body;
        }
      } else {
        $message .= ' - No response received.';
      }

      \Drupal::messenger()->addError($message);
      \Drupal::logger('commerce_ups_fulfillment')->error($message);
    } catch (\Throwable $e) {
      \Drupal::messenger()->addError('An unexpected error has occurred: ' . $e->getMessage());
      \Drupal::logger('commerce_ups_fulfillment')->error($e->getMessage());
    }
    
    return [];
  }

  function attach_image_to_shipment(Shipment $shipment, $base64_string, $filename = 'label.gif'): bool {

    $directory = 'public://shippingLabels'; // Path to save the logo
    $decoded_data = base64_decode($base64_string); // Decode the base64 string to get the raw binary data.

    if ($decoded_data === false) {
      \Drupal::logger('commerce_ups_fulfillment')->error('Failed to decode base64 string.');
      return FALSE;
    }

    // Create Directory if it does not exist
    \Drupal::service('file_system')->prepareDirectory($directory, FileSystemInterface::CREATE_DIRECTORY);

    // Write the file
    $file_uri = $directory . '/' . $filename;
    $file = \Drupal::service('file.repository')->writeData($decoded_data, $file_uri, FileSystemInterface::EXISTS_REPLACE);

    if ($file) {

      // Attach the saved file to the 'field_shipping_label' field.
      $shipment->set('field_shipping_label', [
        'target_id' => $file->id(),
        'alt' => 'Shipping label',
        'title' => 'Label Image',
      ]);
      $shipment->save();

      return TRUE;

    } else {
      \Drupal::logger('commerce_ups_fulfillment')->error('Failed to save file to ' . $file_uri);
      return FALSE;
    }
  }

  /**
   * Prepares the Payload for the UPS Shipment Request
   */
  protected function prepareShipmentRequest(): array {

    $shippingMethod = $this->shipment->getShippingMethod()->getPlugin()->getPluginId();

    if ($shippingMethod === 'ups') {
      $ups_service = $this->shipment->getShippingService();
    } else {
      $ups_service = '03'; // Default to Ground
    }

    $shipment = [];
    $shipment['Description'] = "Timber Shipment";

    $store = $this->shipment->getOrder()->getStore();

    /** @var \CommerceGuys\Addressing\AddressInterface $address_from */
    $address_from = $store->getAddress();

    /** @var \CommerceGuys\Addressing\AddressInterface $address_to */
    $address_to = $this->shipment->getShippingProfile()
      ->get('address')
      ->first();
    
    $company_name = substr($store->getName(), 0, 35);

    $shippingProfile = $this->shipment->getShippingProfile();
    $shipToNumber = '';
    if ($shippingProfile->hasField('field_phone_number') && $shippingProfile->get('field_phone_number')->getValue()) {
      $phoneNumber = $shippingProfile->get('field_phone_number')->getValue();
      $shipToNumber = $phoneNumber[0]['value'];
    }

    $shipFromNumber = '';
    if ($store->hasField('field_phone_number') && $store->get('field_phone_number')->getValue()) {
      $phoneNumber = $store->get('field_phone_number')->getValue();
      $shipFromNumber = $phoneNumber[0]['value'];
    } else {
      // The API Requires a shipFrom Phone Number.
      \Drupal::messenger()->addError('Sender phone number required.  Add a phone number to the Store.');
      \Drupal::logger('commerce_ups_fulfillment')->error('Sender phone number required.  Add a phone number to the Store.');
      return [];
    }

    $addresses = [
      'ShipTo' => $address_to,
      'Shipper' => $address_from,
      'ShipFrom' => $address_from,
    ];
    foreach ($addresses as $direction => $address) {
      $shipment[$direction]['Address'] = [
        'AddressLine' => [
          $address->getAddressLine1(),
          $address->getAddressLine2(),
        ],
        'City' => $address->getLocality(),
        'StateProvinceCode' => $address->getAdministrativeArea(),
        'PostalCode' => $address->getPostalCode(),
        'CountryCode' => $address->getCountryCode(),
      ];
      if (!empty($address->getOrganization())) {
        $shipment[$direction]['AttentionName'] = $address->getOrganization();
      }
    }

    $shipment['ShipTo']['Name'] = $address_to->getGivenName() . ' ' . $address_to->getFamilyName();
    $shipment['ShipTo']['AttentionName'] = $address_to->getGivenName() . ' ' . $address_to->getFamilyName();
    $shipment['ShipTo']['Phone']['Number'] = $shipToNumber;

    $shipment['ShipFrom']['Name'] = $company_name;
    $shipment['ShipFrom']['AttentionName'] = $company_name;
    $shipment['ShipFrom']['Phone']['Number'] = $shipFromNumber;

    $shipment['Shipper']['Name'] = $company_name;
    $shipment['Shipper']['AttentionName'] = $company_name;
    $shipment['Shipper']['Phone']['Number'] = $shipFromNumber;
    $shipment['Shipper']['ShipperNumber'] = $this->config['account_number'];

    $package = $this->getPackage($address_from->getCountryCode());
    $package = reset($package);
    $package['Packaging']['Code'] = $package['PackagingType']['Code'];
    unset($package['PackagingType']);
    $shipment['Package'] = $package;
    $shipment['Service'] = [
      "Code" => $ups_service,
    ];

    $shipment['PaymentInformation'] = [
      "ShipmentCharge" => [
        "Type" => "01",
        "BillShipper" => [
          "AccountNumber" => $this->config['account_number'],
        ],
      ],
    ];

    $request = [
      "ShipmentRequest" => [
        "Request" => [
          "SubVersion" => "1801",
          "RequestOption" => "nonvalidate",
          "TransactionReference" => [
            "CustomerContext" => "",
          ],
        ],
        'Shipment' => $shipment,
      ],
      "LabelSpecification" => [
        "LabelImageFormat" => [
          "Code" => "GIF",
          "Description" => "GIF",
        ],
        "HTTPUserAgent" => "Mozilla/4.5",
      ],
    ];

    return $request;
  }

  /**
   * Returns information about packaging for UPS API.
   *
   * @param string $country_code
   *   The country code for correct measurement.
   */
  protected function getPackage(string $country_code): array {
    $package = [];
    // Set package type.
    $package_type = $this->shipment->getPackageType()
      ?? $this->shippingMethod->getPlugin()->getDefaultPackageType();
    $remote_id = $package_type->getRemoteId();
    $type_code = !empty($remote_id) && $remote_id !== 'custom' ? $remote_id : '00';

    // Set package weight.
    $weight_unit = $country_code === 'US' ? WeightUnit::POUND : WeightUnit::KILOGRAM;
    $weight = $this->shipment->getWeight()->convert($weight_unit);
    $weight_unit = match ($weight->getUnit()) {
      'lb' => 'LBS',
      'kg' => 'KGS',
    };

    // UPS API require weight value to be exact 6 symbols long.
    $weight_number = (string) ceil($weight->getNumber());
    $weight_number_length = strlen($weight_number);
    if ($weight_number_length < 6) {
      for ($i = 0; $i < (6 - $weight_number_length); $i++) {
        $weight_number = '0' . $weight_number;
      }
    }

    // Set package dimensions.
    $length_unit = $country_code === 'US' ? LengthUnit::INCH : LengthUnit::CENTIMETER;
    $length = ceil($package_type->getLength()
      ->convert($length_unit)
      ->getNumber());
    $height = ceil($package_type->getHeight()
      ->convert($length_unit)
      ->getNumber());
    $width = ceil($package_type->getWidth()
      ->convert($length_unit)
      ->getNumber());

    $package[] = [
      'PackagingType' => [
        'Code' => $type_code,
      ],
      'PackageWeight' => [
        'UnitOfMeasurement' => [
          'Code' => $weight_unit,
          'Description' => $weight->getUnit(),
        ],
        'Weight' => $weight_number,
      ],
      'Dimensions' => [
        'UnitOfMeasurement' => [
          'Code' => strtoupper($length_unit),
          'Description' => $length_unit,
        ],
        'Length' => number_format($length, 2, '.'),
        'Height' => number_format($height, 2, '.'),
        'Width' => number_format($width, 2, '.'),
      ],
    ];

    return $package;
  }

  /**
   * Logs information from UPS API requests.
   *
   * @param string $message
   *   The message to log.
   * @param mixed|null $data
   *   Message data.
   *
   * @return void
   */
  protected function log(string $message, $data = NULL): void {
    if ($data) {
      $this->logger->info("@message <br /><pre>@dat</pre>", [
        '@message' => $message,
        '@data' => var_export($data, TRUE),
      ]);
    }
    else {
      $this->logger->info($message);
    }
  }

}